@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: resource.c,v 1.7 2001/03/08 13:59:00 hughesj Exp $
+ *	$Id: resource.c,v 1.11 2004/07/08 19:00:04 roland Exp $
  */
 
 #include "defs.h"
@@ -167,9 +167,9 @@ rlim64_t lim;
 	if (lim == RLIM64_INFINITY)
 		sprintf(buf, "RLIM64_INFINITY");
 	else if (lim > 1024 && lim%1024 == 0)
-		sprintf(buf, "%lld*1024", lim/1024);
+		sprintf(buf, "%lld*1024", (long long) lim/1024);
 	else
-		sprintf(buf, "%lld", lim);
+		sprintf(buf, "%lld", (long long) lim);
 	return buf;
 }
 
@@ -455,6 +455,12 @@ struct tcb *tcp;
 
 #ifdef LINUX
 
+#define NEW_CMD(c)      ((0x80<<16)+(c))
+#define XQM_CMD(c)      (('X'<<8)+(c))
+#define NEW_COMMAND(c) (( ((c) >> SUBCMDSHIFT) & (0x80 << 16)))
+#define XQM_COMMAND(c) (( ((c) >> SUBCMDSHIFT) & ('X' << 8)) == ('X' << 8))
+#define OLD_COMMAND(c) (!NEW_COMMAND(c) && !XQM_COMMAND(c))
+
 static struct xlat quotacmds[] = {
 	{ Q_QUOTAON,	"Q_QUOTAON"	},
 	{ Q_QUOTAOFF,	"Q_QUOTAOFF"	},
@@ -465,6 +471,20 @@ static struct xlat quotacmds[] = {
 	{ Q_SETQLIM,	"Q_SETQLIM"	},
 	{ Q_GETSTATS,	"Q_GETSTATS"	},
 	{ Q_RSQUASH,	"Q_RSQUASH"	},
+	{ NEW_CMD(0x1), "Q_SYNC"        },
+	{ NEW_CMD(0x2), "Q_QUOTAON"     },
+	{ NEW_CMD(0x3), "Q_QUOTAOFF"    },
+	{ NEW_CMD(0x4), "Q_GETFMT"      },
+	{ NEW_CMD(0x5), "Q_GETINFO"     },
+	{ NEW_CMD(0x6), "Q_SETINFO"     },
+	{ NEW_CMD(0x7), "Q_GETQUOTA"    },
+	{ NEW_CMD(0x8), "Q_SETQUOTA"    },
+	{ XQM_CMD(0x1), "Q_XQUOTAON"    },
+	{ XQM_CMD(0x2), "Q_XQUOTAOFF"   },
+	{ XQM_CMD(0x3), "Q_XGETQUOTA"   },
+	{ XQM_CMD(0x4), "Q_XSETQLIM"    },
+	{ XQM_CMD(0x5), "Q_XGETQSTAT"   },
+	{ XQM_CMD(0x6), "Q_XQUOTARM"    },
 	{ 0,		NULL		},
 };
 
@@ -478,10 +498,18 @@ int
 sys_quotactl(tcp)
 struct tcb *tcp;
 {
+	/*
+	 * The Linux kernel only looks at the low 32 bits of the command
+	 * argument, but on some 64-bit architectures (s390x) this word
+	 * will have been sign-extended when we see it.  The high 1 bits
+	 * don't mean anything, so don't confuse the output with them.
+	 */
+	unsigned int cmd = tcp->u_arg[0];
+
 	if (entering(tcp)) {
-		printxval(quotacmds, tcp->u_arg[0] >> SUBCMDSHIFT, "Q_???");
+		printxval(quotacmds, cmd >> SUBCMDSHIFT, "Q_???");
 		tprintf("|");
-		printxval(quotatypes, tcp->u_arg[0] & SUBCMDMASK, "???QUOTA");
+		printxval(quotatypes, cmd & SUBCMDMASK, "???QUOTA");
 		tprintf(", ");
 		printstr(tcp, tcp->u_arg[1], -1);
 		tprintf(", %lu, ", tcp->u_arg[2]);
@@ -491,9 +519,9 @@ struct tcb *tcp;
 
 		if (!tcp->u_arg[3])
 			tprintf("NULL");
-		else if (!verbose(tcp))
+               else if (!verbose(tcp) || !OLD_COMMAND(cmd))
 			tprintf("%#lx", tcp->u_arg[3]);
-                else if (umoven(tcp, tcp->u_arg[3], sizeof(struct dqblk),   
+                else if (umoven(tcp, tcp->u_arg[3], sizeof(struct dqblk),
                     (char *) &dq) < 0)
                         tprintf("???");
 		else {
@@ -551,7 +579,7 @@ struct tcb *tcp;
 		printpath(tcp, tcp->u_arg[0]);
 		tprintf(", ");
 		printxval(quotacmds, tcp->u_arg[1], "Q_???");
-#endif		
+#endif
 		tprintf(", %lu, %#lx", tcp->u_arg[2], tcp->u_arg[3]);
 	}
 	return 0;
